#!/usr/bin/env python
"""
Logger module provides common utilities useful for
setting up logging for Python projects.
"""

import logging
import os

import pali.constants as constants

LOG_LEVEL_NAMES = ['CRITICAL', 'DEBUG', 'ERROR', 'FATAL',
                   'WARN', 'WARNING', 'INFO']
LOG_SETUP_DONE = False


def create_log_dir(log_dir):
    """
    Creates log directory. If not successful raises
    exception after printing message on stdout.
    """
    try:
        if not os.path.exists(log_dir):
            os.makedirs(log_dir)
    except Exception as err:
        print("Failed to create log directory, %s :  %s" %
              (log_dir, str(err)))
        raise err


def set_module_log_level(module_name, log_level):
    """
    Sets the log level for a module. It is particulary
    helpful when third party module's debugging level
    needs to be changed.
    """
    assert log_level in LOG_LEVEL_NAMES, (
        "Invalid log level, %s, requestd for module %s" %
        (log_level, module_name))

    mod_logger = logging.getLogger(module_name)
    mod_logger.setLevel(getattr(logging, log_level))


def setup_logging(log_dir=None, log_file=None, log_level=logging.INFO):
    """
    Sets up logging for the project. Repeated calls are ignored after
    the first call.
    """
    global LOG_SETUP_DONE
    if LOG_SETUP_DONE:
        return  # makes repeated calls idempotent

    log_dir = log_dir or constants.LOG_DIR
    log_file = log_file or constants.LOG_FILE
    log_level = log_level or getattr(logging, constants.LOG_LEVEL)

    create_log_dir(log_dir)  # Create log directory

    log_file_name = os.path.join(log_dir, log_file)

    root_logger = logging.getLogger()
    root_logger.setLevel(log_level)

    file_handler = logging.FileHandler(log_file_name, mode='w')
    file_handler.setFormatter(logging.Formatter(constants.LOG_FORMAT))
    root_logger.addHandler(file_handler)
    LOG_SETUP_DONE = True


def getLogger(module_name):
    """
    Wrapper around logging.getLogger. It helps in making calls route
    through this method so that any common changes can be done here.
    """
    return logging.getLogger(module_name)



#!/usr/bin/env python
'''
This module allows to provide common framework for defining Parameters.

There is a collection of these parameters which is globally accessible. To keep the implementation simple, 
addition of a paramater is not made "Thread Safe" and it is suggested / expected that
all such parameters are defined in the startup of a project in the main thread itself.

It is also advised to add parameters through the exposed interface "add_param" here.
However, these parameters should be consumed through "config" module. It is mainly
done for two reasons. 

- Config files can ovverride parameter values and hence should be used.
- Config files by itself don't define parameter data type. Through Params, type of config value
  can be determined. 
'''

import collections
from pali.logger import getLogger

log = getLogger(__name__)
PARAMS = collections.defaultdict()

class DistType:
    EVEN = 1

class Parameter(object):

    def __init__(self, name, val, val_type=str,
                ab_values = None, ab_enabled = False,
                ab_distribution=DistType.EVEN):
        self.name = name
        self.val_type = val_type
        self.val = val

        assert isinstance(val, val_type), "Invalid val type."

        # If A/B Values are enabled, param's value provided is ignored. 
        self.ab_values = list(ab_values) if ab_values else []
        self.ab_enabled = ab_enabled
        self.ab_distribution = ab_distribution

        self._ab_value_iter = self.get_ab_value()

    def get_ab_value(self):
        for value in self.ab_values:
            yield value
    
    def get_val(self):
        if not self.ab_enabled:
            return self.val
        try:
            return next(self._ab_value_iter)
        except StopIteration:
            self._ab_value_iter = self.get_ab_value()
            return next(self._ab_value_iter)


def add_param(param, val, val_type=str, ab_values = None, ab_enabled = False, ab_distribution='EVEN'):
    """
    Adds Constant in the global constants dictionary.
    """
    PARAMS[param] = Parameter(param, val, val_type, ab_values, ab_enabled, ab_distribution)
    log.debug("Added parameter : %s", param)


def get_param(param):   # DEPRECATED, use config module instead.
    """
    Returns param val.
    """
    return PARAMS[param].get_val()


#!/usr/bin/env python
'''
Scale Tests for WorkerPool.

Various scales tested are:
    - Large number of Tasks (~30K) added and served.
    - Can Queue hold upto 30K tasks at a time (if enough memory)
    - Large number of worker threads
    - Tasks with large runtimes.
    - Can we run 1 Million tasks through a worker pool, assuming
      only 10K Tasks are there at a time in queue.
'''

import unittest

import pali.task as task
import pali.worker as worker

class MyTask(task.Task):

    def __init__(self, index=None):
        """
        A simple task.
        """
        super(MyTask, self).__init__()

        self.index = index
        self.done = False

    def _run(self):
        self.done = True    # mark it done

class TestWorkerPoolScale(unittest.TestCase):

    def test_large_pending_tasks(self):
        """
        Scale test for pushing 3000 tasks on worker pool.
        """
        PENDING_TASK_COUNT = 3000
        tasks = [MyTask(i) for i in range(PENDING_TASK_COUNT)]

        # Push tasks on pool
        thr_count = 20
        tpool = worker.ThreadPool(thr_count)

        # Push tasks on pool but don't start pool yet so as we can
        # test the number of pending tasks. This testcase also highlights
        # the importance of having both the abilities for WorkerPool:
        # a) As context manager as well as b) respective API based invocations.
        for t in tasks:
            tpool.append_task(t)

        # All the tasks must remain in pending queue until pool is
        # started.
        assert(tpool.remaining() == PENDING_TASK_COUNT)

        # Validate that no task is done until it is started..
        self.assertFalse(any(t.done for t in tasks))

        # Test pending tasks status once again.
        assert(tpool.remaining() == PENDING_TASK_COUNT)

        tpool.start()
        tpool.close()

        # Validate that all tasks passed.
        self.assertTrue(all(t.done for t in tasks))

    def test_wp_scale_large_running(self):
        """
        Scale test for pushing 3000 tasks on worker pool.
        """
        PENDING_TASK_COUNT = 3000
        tasks = [MyTask(i) for i in range(PENDING_TASK_COUNT)]

        # Push tasks on pool
        thr_count = 20
        tpool = worker.ThreadPool(thr_count)

        # Push tasks on pool but don't start pool yet so as we can
        # test the number of pending tasks. This testcase also highlights
        # the importance of having both the abilities for WorkerPool:
        # a) As context manager as well as b) respective API based invocations.
        for t in tasks:
            tpool.append_task(t)

        # All the tasks must remain in pending queue until pool is
        # started.
        assert(tpool.remaining() == PENDING_TASK_COUNT)

        # Validate that no task is done until it is started..
        self.assertFalse(any(t.done for t in tasks))

        # Test pending tasks status once again.
        assert(tpool.remaining() == PENDING_TASK_COUNT)

        tpool.start()

        # Once we start, pending tasks will keep getting cleaned up soon.
        # Then we append tasks again on the queue to check that WPool is
        # working fine. Since Task.run doesn't check if task is already
        # run it doesn't error.
        #
        # Add tasks 3 times
        cnt = 3
        while cnt > 0:
            cnt -= 1
            for t in tasks:
                tpool.append_task(t)

        tpool.close()

        # Validate that all tasks passed.
        self.assertTrue(all(t.done for t in tasks))





#!/usr/bin/env python
'''
This module has the example for usecase of Assembly and Pipelines.

This shows a simple example of preparing multiple recipe(s) for a party.
- It assumes there is a menu of (1) Tea, (2) Rice, (3) Pizza and (4) Curry.
- All the 4 dishes on the menu have different recipe so there will be an
  individual pipeline for each dish that executes the recipe in serial order.
- Top level Assembly, "Party", will have these 4 pipelines that can be
  executed in parallel.
'''


import time

from src import pipeline 
from src import logger
from src import task
from src import worker


log = logger.getLogger(__name__)

class DemoStage(pipeline.Stage):

    def run(self, data):
        log.info("Performing : %s ", self.name)
        time.sleep(2)

class MyAssembly(pipeline.Assembly):

    def _run(self):
        with worker.ThreadPool(2) as tpool:
            for pipeline in self.pipelines:
                tpool.append_task(pipeline)

def demo():

    p_tea = pipeline.Pipeline("Tea")
    p_tea.stages.append(DemoStage("Boil Water"))
    p_tea.stages.append(DemoStage("Grate Ginger"))
    p_tea.stages.append(DemoStage("Add Tea Leaves"))
    p_tea.stages.append(DemoStage("Add milk"))
    p_tea.stages.append(DemoStage("Boil Tea for 10 minutes."))


    p_rice = pipeline.Pipeline("Rice")
    p_rice.stages.append(DemoStage("Soak Rice for 10 minutes."))
    p_rice.stages.append(DemoStage("Wash Rice"))
    p_rice.stages.append(DemoStage("Boil Water"))
    p_rice.stages.append(DemoStage("Add Rice"))
    p_rice.stages.append(DemoStage("Boil Rice for 10 minutes."))

    p_pizza  = pipeline.Pipeline("Pizza")
    p_pizza.stages.append(DemoStage("Ferment dough for 4 hours."))
    p_pizza.stages.append(DemoStage("Roll dough."))
    p_pizza.stages.append(DemoStage("Add sauce"))
    p_pizza.stages.append(DemoStage("Add toppings"))
    p_pizza.stages.append(DemoStage("Bake 10 minutes"))
    p_pizza.stages.append(DemoStage("Remove baking plate."))
    p_pizza.stages.append(DemoStage("Bake 15 minutes"))


    # import pdb ; pdb.set_trace()
    # p_tea.run()
    # p_rice.run()
    # p_pizza.run()

    # return

    party = MyAssembly(name='party')
    party.pipelines.extend([p_tea, p_rice, p_pizza])
    party.run()

if __name__ == '__main__':

    demo()



'''
Config module allows to use Python's in-built configparser
for consuming standard configurations for projects. 

Configurations are defined in an .ini file based format.
'''

import configparser
import os
import threading

from pali import constants as constants
from pali.logger import getLogger
from pali import params as params


CFG_MGR = None
CFG_MGR_LOCK = threading.Lock()

log = getLogger(__name__)


class ConfigManager(configparser.ConfigParser):

    def __init__(self, config_file_path=None, param_vals=None):
        super(ConfigManager, self).__init__()

        self.config_file_path = config_file_path or constants.CONFIG_FILE_PATH
        assert os.path.exists(self.config_file_path), "Invalid Config file path"

        self._section = constants.CONFIG_DEFAULT_SECTION

        self.params = param_vals or params.PARAMS if constants.CONFIG_USE_PARAMS else {}

        setattr(self, "optionxform", str)   # Read in case sensitive manner

        self.read(self.config_file_path)

    def read(self, config_path=None):
        """
        Reads config file.
        Also copies common config section in all other sections.
        """
        # determing file to read config from. param has precedence as it can
        # be used to override existing params.
        cfg_file = config_path or self.config_file_path
        super(ConfigManager, self).read(cfg_file)

    def set_section(self, section):
        """
        Changes the current section.
        """
        if self.section == section:
            return
        assert section in self.sections(), "Invalid section : %s" % section
        self._section = section

    @property
    def section(self):
        """
        Returns current default section.
        """
        return self._section

    def get_type_mapping(self, val_type):
        if val_type == int:
            return 'getint'
        elif val_type == float:
            return 'getfloat'
        elif val_type == bool:
            return 'getboolean'
        elif val_type == str:
            return 'get'    # default is str

    def get_param(self, param, val=None, section=None):
        _section = section or self.section

        if param not in self.params:
            return self[_section].get(param, val)

        param_obj = self.params[param]
        mapping = self.get_type_mapping(param_obj.val_type)
        return getattr(self[_section], mapping)(param, param_obj.get_val())

    def set_param(self, param, val, section=None):
        _section = section or self.section
        self[_section][param] = val


def get_config_manager(config_file_path=None):
    """
    Returns Singleton instance of Configuration Manager. 

    Ensures that there are no other instance of configurations
    in the system.
    """
    global CFG_MGR
    if CFG_MGR is None:
        with CFG_MGR_LOCK:
            CFG_MGR = CFG_MGR or ConfigManager(config_file_path)
    return CFG_MGR


def get_param(param , default_val=None, section=None):
    """
    Returns provided or stored value for the param or None if it is not
    set yet.
    """
    return get_config_manager().get_param(param, default_val, section)


def set_param(param, val=None, section=None):
    """
    Sets param and it's value for later use.
    """
    get_config_manager().set_param(param, val, section)



#!/usr/bin/env python
'''
This module defines the Abstract Interface and Implementation for
Pipeline Stages, Pipelines and Assembly.

A Stage is an atomic step as part of whole pipelining process. A Pipeline
is a collection of Stages. An Assembly is collection of Pipelines.

There can be several stages in a pipeline but there must be a serial order
in which they shall be executed. This is the reason Stages are kept in the
form of Linked List inside a pipeline. No two stages of same pipeline can
be executed in parallel. However, there can be multiple pipelines
that can be run in parallel.

A single thread shall be assigned to execute a pipeline stage as it
is always executed in serial order.

Pipelines are nothing but a description of collection of steps to be
executed on a given set of data.

One usecase of pipelines is to configure endpoints / VMs / nodes. Here,
set of steps to be executed is same for all the endpoints but need to be
separately and can also be executed in parallel. In such a case, a pipeline
will be defined once but Assembly will create multiple Pipelines with varying
IP addresses of endpoints to be configured.

'''

import abc

import pali.logger as logger
import pali.task as task
import pali.worker as worker


log = logger.getLogger(__name__)


class Stage(object):

    """
    Stages are like the atomic step of pipelines execution.
    """

    __metaclass__ = abc.ABCMeta
    WAITING = 0x1
    RUNNING = 0x1 << 1
    PASSED = 0x1 << 2
    FAILED = 0x1 << 3

    def __init__(self, name):
        self.name = name

        self._state = self.WAITING
        self._next = None


    def run(self, data=None):
        raise NotImplementedError("Implement run method")


class Pipeline(task.Task):

    def __init__(self, name, stages=None, data=None):
        """
        A pipeline is a collection of pipeline stages.
        """
        super(Pipeline, self).__init__()

        self.name = name
        self.stages = stages if stages else []
        self.data = data if data else {}

    def _run(self):
        log.info("Starting pipeline : %s", self.name)
        for stage in self.stages:
            try:
                stage.run(self.data)
            except Exception as err:
                # TODO : Do better Error handling
                log.error("Pipeline Error : %r", err)

                # if a pipeline stage fais we need to stop the execution.
                break
        log.info("Finished pipeline : %s", self.name)


class Assembly(task.Task):

    MAX_CONCURRENT_PIPELINES = 5

    def __init__(self, name, pipelines=None, max_concurrent_pipelines=None):
        '''
        Assembly is a collection of pipelines.
        '''

        super(Assembly, self).__init__()
        self.name = name
        self.pipelines = pipelines if pipelines else []
        self.max_threads = max_concurrent_pipelines if max_concurrent_pipelines is not None else self.MAX_CONCURRENT_PIPELINES

    def _run(self):

        log.info("Starting Assembly : %s", self.name)
        tpool = worker.ThreadPool(self.max_threads)
        tpool.start()

        for pipeline in self.pipelines:
            tpool.append_task(pipeline)

        tpool.close()
        log.info("Finished Assembly : %s", self.name)



#!/usr/bin/env python
'''
A simple test for config module.
'''
import tempfile
import unittest

from pali import logger as logger
from pali import params as params
from pali import config as config

SAMPLE_CONFIG="""
[DEFAULT]
Compression = yes
ForwardX11 = yes
User = First.Last

[SECTION1]
Port = 50022
ForwardX11 = no
BitRate = 0.9

[SECTION2]
ServerAliveInterval = 45
CompressionLevel = 9
"""

class ConfigTest(unittest.TestCase):

    def test_config(self):
        """
        Tests Params
        """
        params.add_param('ServerAliveInterval', 40, int)
        params.add_param('Compression', False, bool)
        params.add_param('ForwardX11', True, bool)
        params.add_param('BitRate', 1.2, float)

        # Create Config file. 
        with tempfile.NamedTemporaryFile(prefix='test_config') as fp:
            fp.write(bytes(SAMPLE_CONFIG, 'utf-8'))
            fp.flush()

            _ = config.get_config_manager(fp.name)

            # For following tests, default section would be automatically
            # set to "DEFAULT"
            # Test 1 : "Compression"
            # Compression is set to be False in params but in config file
            # it has been enabled by word "yes". Config Manager would read
            # value from config and automatically convert "yes" to True
            assert config.get_param('Compression'), "Compression test failed."

            # Test 2 : "ForwardX11"
            # ForwardX11 is set to be "yes" in DEFAULT section. Though it
            # has been disabled by "no" in section. Config Manager would read
            # value from config and automatically convert "no" to False.
            assert not config.get_param('ForwardX11', section='SECTION1'), \
                "ForwardX11 test failed."

            # Test 3 : Default section values should be available in other
            # sections as well. 
            assert config.get_param('User', section='SECTION1') == "First.Last", \
                "DEFAULT::user not found in DEFAULT SECTION"
            assert config.get_param('User', section='SECTION2') == "First.Last", \
                "DEFAULT::user not found in SECTION1 SECTION"



if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
'''
This module contains the example of VM deployment through ovftool command.

A typical VM deployment can take anywhere 2-20 minutes depending on the size
of VM. This becomes a real good example of demonstrating that threads in the
Thread Pool are waiting for the tasks to be pushed in the Queue and take the
task from queue as soon as they become available which means they don't wait
for all the tasks to be pushed and "exeucte" or "run" to be called for tasks
to be picked up.
'''

import subprocess
import pipes


from src import logger
from src import task
from src import worker


log = logger.getLogger(__name__)

class OvfTask(task.Task):
    TEMPLATE=''
    INFRAHOST_IP=''
    DATASTORE=''
    PASSWORD=''
    OVF_CMND='ovftool --acceptAllEulas --datastore=%s --noSSLVerify --name=%s %s vi://root:%s@%s'

    def __init__(self, name):
        super(OvfTask, self).__init__()
        self.name = name

    def run_command(self, cmd):
        """
        Helper routing for running commands on command line.
        """
        log.debug("Running command: %s", " ".join(cmd))
        p = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        stdout, stderr = p.communicate()
        if p.returncode != 0:
            msg = "command %r failed, stdout: %r, stderr: %r"
            raise RuntimeError(msg % (" ".join(cmd), stdout, stderr))
        return stdout.rstrip().decode()

    def get_command(self):
        """
        Builds and returns the ovftool command.
        """
        return self.OVF_CMND % (self.DATASTORE,
                                self.name, self.TEMPLATE,
                                self.PASSWORD, self.INFRAHOST_IP)

    def _run(self):
        try:
            cmnd = self.get_command()
            log.info("Deploying VM %s through command : %s", self.name, cmnd)
            out = self.run_command(cmnd.split())
            log.info("Deployment finished for VM %s", self.name)
        except Exception as err:
            log.info("Deployment failed for VM %s", self.name)


def demo():

    tasks = [OvfTask(name='vip_esx_vm_%s' % x) for x in range(4)]

    # Add tasks one by one to highlight that worker threads are waiting for
    # jobs to be pushed onto the queue and will pickup the job as soon as
    # it becomes available on the queue.
    with worker.ThreadPool(2) as tpool:
        for t in tasks:
            tpool.append_task(t)

if __name__ == '__main__':
    demo()



#!/usr/bin/env python
'''
A simple test to check pause on worker threads.

To run the test (from repo root directory) :
    python -m unittest discover -p test_bbuffer.py ./tests
'''
import time
import unittest

import pali.logger as logger
from examples.bbuffer import TimeStampRecorder, Producer

log = logger.getLogger(__name__)


class TestWorkerPoolPause(unittest.TestCase):

    def test_thread_task_loop(self):
        pc = Producer()
        
        # Test : Start thread and wait for 3 seconds. 
        # Records should be peoduced in the buffer.
        pc.start()
        time.sleep(3)
        pcount = pc.produce_count()
        assert pcount, "Threaded Task didn't start"

        # Test : Stop thread and wait for 2 seconds.
        # Take produce count and wait for another 4 seconds, 
        # No more records should have been produced in the buffer
        pc.stop()
        time.sleep(2)
        pcount = pc.produce_count()
        time.sleep(4)
        assert pc.produce_count() == pcount, "Threaded Task didn't stop"


    def test_bounded_buffer(self):
        tc = TimeStampRecorder()

        # Test  Producers can be started / stopped.
        # Test : Start thread and wait for 3 seconds. 
        # Records should be peoduced in the buffer.
        tc.start(consumers=False)   # start prodcuers only.
        time.sleep(3)
        pcount = tc.get_buffer_size()
        assert pcount, "Timestamp Producer didn't start"

        # Test : Stop thread and wait for 2 seconds.
        # Take produce count and wait for another 4 seconds, 
        # No more records should have been produced in the buffer
        tc.stop(consumers=False)    # start producer only
        time.sleep(2)
        pcount = tc.get_buffer_size()
        time.sleep(4)
        assert tc.get_buffer_size() == pcount, "Timestamp Producer didn't stop"



        # Test : Consumers can be started / stopped.
        tc.start(producers=False)   # start consumers only.
        time.sleep(10)
        pcount = tc.get_buffer_size()
        assert not pcount, "Timestamp Consumer didn't start"

        # Test : Stop thread and wait for 2 seconds.
        # Take produce count and wait for another 4 seconds, 
        # No more records should have been produced in the buffer
        tc.stop(producers=False)    # start consumers only.
        tc.start(consumers=False)   # start producer again.
        time.sleep(4)
        tc.stop(consumers=False)   # stop producer again.
        pcount = tc.get_buffer_size()
        time.sleep(3)
        assert pcount == tc.get_buffer_size(), "Timestamp Consumer didn't stop"

        
       


#!/usr/bin/env python
'''
Some general purpose utilities.
'''

import platform

PYTHON_VER = platform.python_version()

PYTHON_2 = PYTHON_VER.startswith('2')
PYTHON_3 = PYTHON_VER.startswith('3')

system = platform.system()
IS_WINDOWS = system == 'Windows'
IS_LINUX = system == 'Linux'
IS_MAC = system == 'Darwin'

if PYTHON_2:
    import Queue as queue
elif PYTHON_3:
    import queue as queue


queue = queue



#!/usr/bin/env python
'''
Worker Pool testcases.
'''

import unittest

import pali.task as task
import pali.worker as worker


class MyTask(task.Task):

    def __init__(self, index=None):
        """
        A simple task.
        """
        super(MyTask, self).__init__()

        self.index = index
        self.done = False

    def _run(self):
        self.done = True    # mark it done


class TestWorkerPool(unittest.TestCase):

    def test_worker_pool(self):
        tasks = [MyTask(i) for i in range(5)]

        # Push tasks on pool
        thr_count = 2
        tpool = worker.ThreadPool(thr_count)
        tpool.start()
        for t in tasks:
            tpool.append_task(t)
        tpool.close()

        # Validate that all tasks passed.
        self.assertTrue(all(t.done for t in tasks))

    def test_context_manager(self):
        """
        Same as above test but test context manager
        """
        tasks = [MyTask(i) for i in range(5)]

        # Push tasks on pool
        thr_count = 2

        # Invoke WorkerPool as in context manager.
        # Note that .start() and .close() w.r.t. to
        # previous test have been moved to context.
        with worker.ThreadPool(thr_count) as tpool:
            _ = [tpool.append_task(t) for t in tasks]

        # Validate that all tasks passed.
        self.assertTrue(all(t.done for t in tasks))

    def test_task_order(self):
        """
        Tests that tasks are processed in same order. Uses ThreadPool
        in single Thread mode to enusre that tasks are processed in
        the order they are pushed.

        This test case MUST pass for ThreadPool to be usable in
        building data pipelines.
        """
        class LinkedTask(task.Task):
            def __init__(self, index):
                self._id = index
                self.done = False
                self.prev = None

            def _run(self):
                # Need not check all previous nodes to be marked
                # done here. "Done" is a Transitive property here
                # and that means if aRb and bRc then aRc.
                assert(not self.prev or self.prev.done)
                self.done = True

        # Forms a linked List of Tasks and puts them on worker pool.
        # When each task is run it checks , all of its previous are
        # marked as done.
        TASK_COUNT = 10
        prev_task, tasks = None, []
        for i in range(TASK_COUNT):
            t = LinkedTask(i)
            tasks.append(t)
            t.prev, prev_task = prev_task, t

        # Push tasks on pool
        thr_count = 1

        # Invoke WorkerPool as in context manager.
        # Note that .start() and .close() w.r.t. to
        # previous test have been moved to context.
        with worker.ThreadPool(thr_count) as tpool:
            _ = [tpool.append_task(t) for t in tasks]

        # Validate that all tasks are marked done.
        # A node will be marked done only if all its
        # previous nodes in the list were marked done.
        self.assertTrue(all(t.done for t in tasks))


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
'''
This module provides core implementation of ThreadPool.
'''

import threading as threading

from pali.common import queue
import pali.logger as logger

log = logger.getLogger(__name__)

class WorkerThread(threading.Thread):

    def __init__(self, in_queue, out_queue, name=None, verbose=False):

        threading.Thread.__init__(self)

        self._input_queue = in_queue
        self._output_queue = out_queue
        self._verbose = verbose

        # Thread control mechanism
        self._stop_event = threading.Event()
        self._stop_event.clear()

        # Set daemon thread, since run method does
        # blocking wait on items from queue
        self.daemon = True

    def run(self):
        while not self._stop_event.is_set():
            try:
                if self._verbose:
                    t = threading.current_thread()
                    log.debug(" Thread %s : popping element", t.name)

                # This is a blocking call.
                task = self._input_queue.get()

                if self._verbose:
                    log.debug(" Thread %s : popped element", t.name)

                # We handle only task.Task based requests.
                # assert(isinstance(task, task.Task))

                self.run_task(task)

                self._input_queue.task_done()

            except Exception as err:
                log.exception(err)
                pass

        log.info("Finished execution.")


    def run_task(self, task):
        try:
            task.run()
        except Exception as err:
            log.exception("Exception in rrunning task %s , %r", task, err)
            pass

    def stop(self):
        self._stop_event.set()

    disable = stop

    def enable(self):
        self._stop_event.clear()

    def is_disabled(self):
        """
        Returns True if worker thread is disabled.
        """
        return self._stop_event.isSet()

    def close(self):
        # Stop the thread.
        self.stop()
        # TODO : What else do we want to do.


class WorkerPool(object):
    MAXSIZE = 3000
    MAX_PARALLEL_TASK = 10

    def __init__(self, max_parallel=None, max_queue_size=None):
        '''
        A worker pool that simply takes job from pending tasks and
        assigns to one of the worker threads.
        '''
        maxsize = max_queue_size if max_queue_size else WorkerPool.MAXSIZE
        self._max_parallel_tasks = max_parallel if max_parallel else self.MAX_PARALLEL_TASK

        self._pending_tasks = queue.Queue(maxsize=maxsize)
        self._finished_tasks = queue.Queue(maxsize=maxsize)

        self._handlers = []

    def append_task(self, task):
        try:
            self._pending_tasks.put(task)
        except Exception as err:
            log.exception("Exception in adding task to queue %r", err)

    def close(self):
        self._pending_tasks.join()
        for handler in self._handlers:
            handler.stop()

    def remaining(self):
        """Returns (approx.) number of pending tasks."""
        return self._pending_tasks.qsize()

    def finished(self):
        return self._finished_tasks.qsize()

    def get_workers(self):
        '''
        Returns the list of worker threads.
        '''
        return self._handlers


class ThreadPool(WorkerPool):

    def __init__(self, max_threads=None, max_queue_size=None, verbose=False):
        self._verbose = verbose
        super(ThreadPool, self).__init__(max_parallel=max_threads,
                                         max_queue_size=max_queue_size)
        self.initialize()

    def initialize(self):
        # Initialize the handlers.
        self._handlers = [ WorkerThread(in_queue=self._pending_tasks,
                                        out_queue=self._finished_tasks,
                                        verbose=self._verbose)
                                    for i in range(self._max_parallel_tasks) ]

    def start(self):
        for handler in self._handlers:
            handler.start()

    def __enter__(self):
        self.start()
        return self

    def __exit__(self, _type, value, traceback):
        self.close()



#!/usr/bin/env python
'''
A simple test for the task.
'''
import unittest

from pali import logger as logger
from pali import params as params


class ParamsTest(unittest.TestCase):

    def check_param_value(self, param, values):
        # TEST 2 : Test if values iterate for A/B enabled.
        params.add_param(param=param, val=False,
                         val_type=bool, ab_values =values,
                         ab_enabled=True)

        # Try to get value of param for 10 times and
        # it should simply iterate over values. 
        for _ in range(10):
            assert params.get_param(param) in values, \
                "A/B values not in values provided."

    def test_params(self):
        """
        Tests Params
        """
        # TEST 1 : Check if value is added properly.
        params.add_param('key', 'value')
        assert params.get_param('key') == 'value'

        # TEST 2 : Check for bool param values.
        self.check_param_value('enabled', [True, False])

        # TEST 3 : Check for int param values.
        self.check_param_value('int_vals', list(range(2, 5)))

        # TEST 4 : Check for int param values.
        self.check_param_value('int_vals', list(range(6)))

        # TEST 5 : Check for float param values.
        self.check_param_value('float_vals', [10.3, 4.5, 6.2, 3.56])


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
"""
This module defines a wrapper for running child processes using
subprocess module.
"""


import os
import shlex
import signal
import subprocess
import time

from pali import logger
from pali import common

log = logger.getLogger(__name__)


class Console(object):
    """
    A wrapper around subprocess handle that does bookkeeping for the
    returned handle. Also provides necessary utility functions for
    checking state of child process or killing it in a proper manner.
    """

    def run_command(self, cmnd, env=None, cwd=None, timeout=-1):
        """
        Runs "cmnd" in single string and returns a tuple of  return code and
        result.

        Parameters
        ----------
        cmnd : str
            Command to be executed.
        env : string
            Set of variable to be set in string.
        cwd : string
            Working directory for child process.
        timeout : int
            time (in secs) to wait for child process to finish execution.
            -1 as default sets no time limit.

        Returns
        -------
        (int, string)
            returns a tuple of command execution return code and output.
        """

        self.proc = self._create_subprocess(cmnd=cmnd, cwd=cwd, env=env)

        # Start the counter for command to finish if requested.
        if timeout > 0:
            time_limit = time.time() + timeout

            while time.time() < time_limit:
                if self.is_alive(self.proc):
                    time.sleep(1)
                else:
                    break
            self.kill_process(self.proc)

        stdout_val = self.proc.communicate()[0]
        return self.proc.returncode, stdout_val.strip().decode('utf-8')

    def _create_subprocess(self, cmnd, stdout=None, stderr=None,
                           env=None, cwd=None):
        """
        Starts a subprocess and returns a handle for it.
        """
        stdout = subprocess.PIPE if not stdout else stdout
        stderr = subprocess.STDOUT if not stderr else stderr

        cmnd = shlex.split(cmnd)
        if common.is_windows():
            return subprocess.Popen(cmnd, shell=False, cwd=cwd, env=env,
                                    stdout=stdout, stderr=stderr,
                                    close_fds=False, start_new_session=True)
        else:
            return subprocess.Popen(cmnd, shell=False, cwd=cwd, env=env,
                                    stdout=stdout, stderr=stderr,
                                    close_fds=True, preexec_fn=os.setsid,
                                    bufsize=-1)

    def ctrl_c(self, timeout=1):
        """
        Sends SIGINT(Ctrl+C) interrupt to child process.
        """
        self.proc.send_signal(signal.SIGINT)
        # Wait for some time after sending SIGINT signal :
        # https://bugs.python.org/issue25942
        self.proc.wait(timeout)

    def kill_process(self, close_fds=False):
        """
        Kills a subprocess represented by handle 'proc'.
        """
        # command is still active, kill it.
        log.warning("Killing child process %s" % self.proc.pid)
        self.proc.kill()
        if close_fds:
            self.proc.communicate()

    def is_alive(self):
        """
        Returns True if subprocess is still running else False
        """
        return self.proc.poll() is None



#!/usr/bin/env python
'''
A simple test for the task.
'''

import pali.task as task
import unittest


class MyTask(task.Task):
    def _run(self):
        pass


class FirstTest(unittest.TestCase):

    def setUp(self):
        self._task = MyTask()

    def test_first(self):
        """
        A simple test to check Task can be extended and
        executed.
        """
        self._task.run()


if __name__ == '__main__':
    unittest.main()



'''
This module implements Bounded Buffer (a.k.a. Producer Consumer).
'''
from pali.common import queue
from pali.thread import ThreadTaskLoop

from pali import logger

log = logger.getLogger(__name__)

class ProducerConsumer(object):
    BUFFER_SIZE = 1000
    RECORD_PUT_TIMEOUT = 2
    RECORD_GET_TIMEOUT = 2
    NAME = ''

    class Producer(ThreadTaskLoop):
        pass

    class Consumer(ThreadTaskLoop):
        pass
    def __init__(self, buffer_size=None):
        """
        ProducerConsumer defines a general Producer Consumer pattern
        that can be used by various applications by simply defining
        "produce" and "consume" methods.

        Parameters
        -----------
        buffer_size : int
            Size of the buffer that holds elements produced by producer.
        """
        self.buffer_size = buffer_size or self.BUFFER_SIZE
        self._buffer = queue.Queue(maxsize=self.buffer_size)

        self._producer = self.Producer()
        self._producer._task = self.produce

        self._consumer = self.Consumer()
        self._consumer._task = self.consume

    @property
    def cname(self):
        return self.NAME or self.__class_.name

    def produce(self):
        """
        Produces elements and adds them in buffer.
        """
        raise NotImplementedError("%s::produce is not defined", self.cname)

    def consume(self):
        """
        Takes elements from buffer and processes them.
        """
        raise NotImplementedError("%s::produce is not defined", self.cname)

    def start(self, producers=True, consumers=True):
        """
        Starts producer and consumer tasks in separate threads.

        Parameters
        ----------
        producers: bool
            starts producer thread if set. Defaults to True
        consumers: bool
            starts consumer thread if set. Defaults to True
        """
        if producers:
            self._producer.start()

        if consumers:
            self._consumer.start()

    def stop(self, producers=True, consumers=True):
        """
        Stops producer and consumer tasks in separate threads.

        Parameters
        ----------
        producers: bool
            stops producer thread if set. Defaults to True
        consumers: bool
            stops consumer thread if set. Defaults to True
        """
        if producers:
            self._producer.stop()

        if consumers:
            self._consumer.stop()

    def add_to_buffer(self, record):
        """
        Adds record to buffer with FIFO order.

        Parameters
        ----------
        record : Any
            Adds record to internal buffer if buffer is not full already.
            If buffer is full, an error message is generate and record is
            ignored.
        """
        try:
            self._buffer.put(record, block=False,
                             timeout=self.RECORD_PUT_TIMEOUT)
        except queue.Full:
            log.error("Queue full %s", self.cname)
        except Exception as err:
            log.error("Error : %s", err)
            raise

    def pop_from_buffer(self):
        """
        Returs record from buffer in FIFO order.

        Parameters
        ----------
        record : Any
            Adds record to internal buffer if buffer is not full already.
            If buffer is full, an error message is generate and record is
            ignored.
        """
        try:
            return self._buffer.get(block=False, timeout=self.RECORD_GET_TIMEOUT)
        except queue.Empty:
            log.debug("Queue full %s", self.cname)
        except Exception as err:
            log.error("Error : %s", err)
            raise

    def get_buffer_size(self):
        """
        Returns approximate buffer size.
        """
        return self._buffer.qsize()


#!/usr/bin/env python
'''
This module provides utilities for parallelizing work.
'''
from pali.common import PYTHON_2

if PYTHON_2:
    raise ImportError('"parallel" module is available only Python3.')

from concurrent.futures import ThreadPoolExecutor, as_completed

from pali.config import get_param
from pali.logger import getLogger

log = getLogger(__name__)

THREAD_COUNT = get_param('THREAD_POOL_CONCURRENCY', 32)


def ThreadPool(params, blocking=True, thread_count=None):
    """
    A wrapper method around python's in-built ThreadPool executor.

    Parameters
    ------------
    params : List
        List of tuples to be run in parallel in a threadpool.
    blocking : bool
        Waits for all tasks to finish when set to True else
        returns immediately after submission of tasks on Threadpool.
    thread_count: int
        Number of threads to be used in Thread Pool.

    Return
    ------------
    results : dict
        A dictionary of result of each task.
    """
    results = {}
    max_workers = thread_count or THREAD_COUNT

    with ThreadPoolExecutor(max_workers=max_workers) as tpool:
        futures = {}
        for _func, _ident, _args, _kwargs in params:
            _tfunc = tpool.submit(_func, *_args, **_kwargs)
            futures[_tfunc] = _ident

        if not blocking:
            return None

        for future in as_completed(futures):
            ident = futures[future]
            try:
                results[ident] = future.result()
            except Exception as err:
                log.warn("Error for %s - %s", ident, err)
    return results



#!/usr/bin/env python
'''
This module holds package level constants for Pali.
'''


# Config module constants
CONFIG_FILE_PATH = '/etc/pali/pali.cfg'
CONFIG_DEFAULT_SECTION = 'DEFAULT'
CONFIG_COMMON_SECTION = 'COMMON'
CONFIG_USE_PARAMS = True

# Logger constants
LOG_DIR = './'
LOG_FILE = 'pali.log'
LOG_FORMAT = (
    '%(asctime)s::%(levelname)s::%(threadName)s::'
    '%(module)s[%(lineno)04s]::%(message)s')
LOG_LEVEL = 'INFO'



#!/usr/bin/env python
'''
A simple test to check pause on worker threads.

To run the test (from repo root directory) :
    python -m unittest discover -p test_tp_workers.py ./tests
'''
import collections
import threading
import time
import unittest

import pali.task as task
import pali.worker as worker
import pali.logger as logger

log = logger.getLogger(__name__)

class MyTask(task.Task):

    def __init__(self, index, exec_bk, sleep=1):
        """
        A simple task.
        """
        super(MyTask, self).__init__()

        self.index = index
        self.done = False
        self._sleep = sleep
        self.exec_bk = exec_bk
        self.executed_on = None

    def _run(self):
        log.info("Starting task %d", self.index)
        self.done = True    # mark it done
        self.executed_on = threading.current_thread().name
        self.exec_bk[self.index] = self.executed_on
        time.sleep(self._sleep)
        log.info("Finished executing task %d", self.index)


class TestWorkerPoolPause(unittest.TestCase):

    def test_thread_stop(self):
        """
        Check ability to stop and pause the threads. 
        """
        task_thr_map = {}   # thread on which task executed map

        # Generate 20 tasks
        tasks = [MyTask(i,task_thr_map,i) for i in range(7)]

        # list_iter = lambda x : yield n for n in x
        def list_iter(_list):
            for elem in _list:
                yield elem
        
        tasks_iter = list_iter(tasks)

        with worker.ThreadPool(3) as tp:
            # Push first 3 tasks on thre threads. Sleep of 3 secs.
            # ensures that each thread gets one task.
            tp.append_task(next(tasks_iter))
            tp.append_task(next(tasks_iter))
            tp.append_task(next(tasks_iter))

            # stop 2 threads
            thr_act = [False, True, True]
            for worker_thread, stop in zip(tp.get_workers(), thr_act):
                if stop:
                    # perform lazy stop
                    worker_thread.stop()

            # Pausing thread through stop on individual worker thread does
            # lazy-pause which means it will eventually pause and not take
            # more than one item from the queue.

            # Push next 4 tasks on Threadpool just to flush the queue.
            tp.append_task(next(tasks_iter))
            tp.append_task(next(tasks_iter))
            tp.append_task(next(tasks_iter))
            tp.append_task(next(tasks_iter))

            # Wait for all the tasks pushed so far to finish. 
            while not all(tasks[i].done for i in range(7)):
                time.sleep(3)

            log.info(" Task Thread map (before): %r", task_thr_map)

            # Now push all the tasks again on the queue
            task_thr_map = {}
            for t in tasks:
                t.done = False  # mark the tasks undone
                tp.append_task(t) # push on queue again

            # Wait for all the tasks pushed so far to finish. 
            while not all(tasks[i].done for i in range(7)):
                time.sleep(3)

            thr_name = tp.get_workers()[0].name # get first/active thread name
            assert all(t.executed_on == thr_name for t in tasks)



'''
A simple python module to scan Stock Ticker symbols.
'''

import datetime
import json
import logging
import requests


from src import logger 
from src import task
from src import worker


log = logger.getLogger(__name__)
logger.getLogger("requests").setLevel(logging.WARNING)
logger.getLogger("urllib").setLevel(logging.WARNING)

class StockTask(task.Task):
    FIN_DATA_URL='https://www.alphavantage.co'
    API_KEY='ADD_YOUR_KEY'
    FIN_DATA_TYPE='TIME_SERIES_DAILY_ADJUSTED'

    def __init__(self, tckr='TSLA'):
        super(StockTask, self).__init__()
        self.tckr = tckr
        self.data = None
        self.date = datetime.datetime.now().strftime("%Y-%m-%d")

    def build_url(self):
        self.url = '%s/query?apikey=%s&function=%s&symbol=%s' % (
                self.FIN_DATA_URL,
                self.API_KEY,
                self.FIN_DATA_TYPE,
                self.tckr)

    def _run(self):
        try:
            self.build_url()
            r = requests.get(self.url)
            if r.status_code == 200:
                self.data = json.loads(r.text)
                opening = self.open_today()
                log.info("%s opened at : %s", self.tckr, opening)
        except Exception as err:
            log.exception("Tckr: %s , data: %s", self.tckr, self.data)

    def open_today(self):
        if not self.data:
            return None

        return self.data["Time Series (Daily)"][self.date]["1. open"]

def simple():


    data = ['VMW', 'AAPL', 'GOOG', 'TSLA', 'CRWD'] 
    tasks = [StockTask(tckr=x) for x in data]
    with worker.ThreadPool(2) as tpool:
        _ = [tpool.append_task(t) for t in tasks]

    open_today = [(x.tckr, x.open_today()) for x in tasks if x.data]

    log.info("Open today : %s", open_today)

def get_next_tckr():
    alphas = [chr(i).upper() for i in range(97, 97 + 26)]

    for i in alphas + ['_']:
        for j in alphas:
            for k in alphas:
                for l in alphas:
                    tckr = j + k + l
                    tckr = tckr if i == '_' else i + tckr
                    yield tckr
def scrapper():
    log.info("======================")
    log.info("    Start Scrapping  ")
    log.info("======================")
    with worker.ThreadPool(max_threads=15, max_queue_size=26**4) as tpool:
        for tckr in get_next_tckr():
            # log.info("Appending task for : %s", tckr)
            task = StockTask(tckr)
            tpool.append_task(task)


if __name__ == '__main__':
    simple()



#!/usr/bin/env python

import threading

from pali.common import queue
import pali.logger as logger
import pali.task as task
import pali.worker as worker


log = logger.getLogger(__name__)

class MyTask(task.Task):

    def __init__(self, index):
        self._id = index

    def _run(self):
        pass

soln = {}

def _worker():
    while True:
         item = q.get()
         # do_work(item)
         tid = threading.current_thread().ident
         x = soln.get(tid, [])
         x.append(item)
         soln[tid] = x
         q.task_done()

def test_simple():
    q = queue.Queue()
    num_worker_threads = 3
    thr = []
    for i in range(num_worker_threads):
        # t = threading.Thread(target=worker)
        # t.daemon = True
        t = worker.WorkerThread(in_queue=q, out_queue=None, verbose=False)
        thr.append(t)
        t.start()

    for i in range(10):
        item = MyTask(i)
        q.put(item)

    q.join()
    for t in thr:
        t.stop()
        q.put(None)
    log.info("Solution : %s", soln)

if __name__ == '__main__':
    test_simple()



'''
This module :
1. defines wrapper over Python's threading.Thread class.
2. defines ThreadTaskLoop that runs a task in loop.
'''

import threading

from pali.common import PYTHON_2, PYTHON_3
import pali.logger as logger

log = logger.getLogger(__name__)


class Thread(threading.Thread):
    """
    Wrapper over Python's Thread class with daemon property set to
    True by default.

    NOTE: Python allows you to override only __init__ and run method
    for this class.
    """

    def __init__(self, daemon=True, *args, **kwargs):
        # 'daemon' attribute on Thread object is accepted in constructor
        # argument from Python 3 onwards only. In Python 2, it was not
        # accepted as argument but was rather set as attribute on object.
        if PYTHON_3:
            kwargs['daemon'] = daemon

        super(Thread, self).__init__(*args, **kwargs)

        if PYTHON_2:
            self.daemon = daemon


class ThreadTaskLoop(object):
    THREAD_JOIN_TIMEOUT = 3

    def __init__(self, init_switch=True, daemon=True, *args, **kwargs):
        """
        ThreadTask runs a task (defined by _task method) in a loop on a
        separate thread.

        Parameters
        ------------
        init_switch
        """
        self.daemon = daemon
        self._task_switch = threading.Event()
        self._task_thread = None
        if init_switch:
            self.switch_on()

    def _task(self):
        raise NotImplementedError("_task is not implemented.")

    def switch_on(self):
        """
        Turns on task switch.
        """
        self._task_switch.set()
    
    def switch_off(self):
        """
        Turns off task switch.
        """
        self._task_switch.clear()

    def is_switch_on(self):
        """
        Returns True if task switch is ON , False otherwise.
        """
        return self._task_switch.is_set()

    def task(self):
        """
        Runs task in a loop.
        """
        while self.is_switch_on():
            self._task()

    def start(self):
        """
        Start task on a separate thread.
        """
        if self._task_thread:
            log.error("Task thread already started.")
            return

        try:
            self._task_thread = Thread(target=self.task, daemon=self.daemon)
            self._task_thread.start()
        except Exception as err:
            log.error("Error in starting task : %r", err)

    def stop(self):
        """
        Stops the thread running the task and deletes object.
        """
        if not self._task_thread:
            log.error("Task is not running for %s ", self.__name__)
            return

        try:
            self.switch_off()
            self._task_thread.join(timeout=self.THREAD_JOIN_TIMEOUT)
        except Exception as err:
            log.error("Error in stopping task : %r", err)
        finally:
            del self._task_thread
            self._task_thread = None



#!/usr/bin/env python

import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

version = {}
with open("pali/__init__.py") as fp:
    exec(fp.read(), version)

setuptools.setup(
    name="pali",
    version=version['__version__'],
    author="Vipin Sharma",
    author_email="sh.vipin@gmail.com",
    description="A simple ThreadPool library",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/gitvipin/pali",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 2",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)



#!/usr/bin/env python
'''
This module represents the Basic Task in Pali system.


A Task has a well defined life cycle and is treated accordingly. All
the Tasks in the Pali system MUST be derived from it. This class in
itself is non-instantiable with the help of abstract method 'run' and
'_run'.

A Task object should be self contained and it's run method shouldn't
accept any arguments. Any attribute, metadata, configuration or
properties needed for the task must be passed in the contrsuctor at the
time of initialization.
'''
import abc

class Task(object):
    (NEW, READY, RUNNING, WAITING, FINISHED) = range(5)
    __metaclass__ = abc.ABCMeta

    DEFAULT_PRIORITY = 1

    def __init__(self, priority=None):
        """
        This class defines the abstract representation of Task.
        """
        self._state = self.NEW
        self._priority = priority if priority else self.DEFAULT_PRIORITY
        super(Task, self).__init__()

    @property
    def state(self):
        return self._state

    @state.setter
    def state(self, val):
        # TODO : do validation
        self._state = val

    @state.deleter
    def state(self):
        del self._state

    def run(self):
        """
        This method exposes the interface for running the task. This is a
        non overridable method in class so as derived class types do not
        override the default behavior of execution of tasks in the system.
        """
        try:
            self._run()
        except Exception as err:
            # TODO: Do Task Failure to run exception handling
            pass

    @abc.abstractmethod
    def _run(self):
        """
        A helper 'run' routine for running Tasks in Pali system. The core
        functionality of task will be run through this helper routine.
        """
        raise NotImplemented("Abstract method '_run' need to be defined")

    def __lt__(self, other):
        """
        A basic comparator function that compares tasks on their priority.
        """
        return self._priority < other._priority



'''
This module demonstrates examples of ProducerConsumer and
ThreadTaskLoop
'''
import time

from pali.bbuffer import ProducerConsumer
from pali.logger import getLogger, setup_logging
from pali.thread import ThreadTaskLoop

log = getLogger(__name__)

class TimeStampRecorder(ProducerConsumer):
    """
    A simple application of Bounded Buffer a.k.a ProducerConsumer.
    Producer generates timestamps every 2 seconds. 
    Consumer simply consumers timestamps generated and prints them on
    stdout.

    """
    NAME = "TimestampRecorder"

    def produce(self):
        """Generate timestamp every 2 seconds"""
        time.sleep(1)
        self.add_to_buffer(int(time.time()))

    def consume(self):
        """ Print timestamp every 1 second from buffer"""
        time.sleep(2)
        data = self.pop_from_buffer()
        log.info("Data : %s - Buffer size : %s", data,
                 self.get_buffer_size())


class Producer(ThreadTaskLoop):
    """
    A simple ThreadTaskLoop that will run '_task' method
    in a loop.
    """
    def __init__(self):
        super(Producer, self).__init__()
        self.array = []

    def produce_count(self):
        """ Returns number of elements produced so far."""
        return len(self.array)

    def _task(self):
        time.sleep(1)
        self.array.append(int(time.time()))


if __name__ == '__main__':
    setup_logging()

    # Create an instance of Producer and once started, 
    # We should see timestamps collected in it's
    # internal storage.
    pc = Producer()
    pc.start()

    # Create a timestamp recorder
    tc = TimeStampRecorder()
    tc.start()
    time.sleep(10)
    tc.stop()